#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = ["Ed2kHash", "ed2k_hash", "ed2k_hash_async"]
__version__ = (0, 0, 1)

from typing import Final

from filewrap import (
    Buffer, SupportsRead, 
    bio_chunk_iter, bio_chunk_async_iter, 
    bytes_to_chunk_iter, bytes_to_chunk_async_iter, 
)
from Crypto.Hash.MD4 import MD4Hash


MD4_BLOCK_SIZE: Final = 1024 * 9500
MD4_EMPTY_HASH: Final = b"1\xd6\xcf\xe0\xd1j\xe91\xb7<Y\xd7\xe0\xc0\x89\xc0"


class Ed2kHash:

    def __init__(self, b: Buffer = b"", /):
        self.block_hashes = bytearray()
        self._last_hashobj = MD4Hash()
        self._remainder = 0
        self.update(b)

    def update(self, b: Buffer, /):
        size = len(b)
        if not size:
            return
        block_hashes = self.block_hashes
        remainder = self._remainder
        last_hashobj = self._last_hashobj
        m = memoryview(b)
        start = 0
        if remainder:
            start = MD4_BLOCK_SIZE - remainder
            last_hashobj.update(m[:start])
            block_hashes[-16:] = last_hashobj.digest()
        if start < size: 
            for start in range(start, size, MD4_BLOCK_SIZE):
                last_hashobj = MD4Hash(m[start:start+MD4_BLOCK_SIZE])
                block_hashes += last_hashobj.digest()
        self._remainder = (size - start) % MD4_BLOCK_SIZE
        self._last_hashobj = last_hashobj

    def digest(self, /) -> bytes:
        block_hashes: Buffer = self.block_hashes
        if not self._remainder:
            block_hashes = block_hashes + MD4_EMPTY_HASH
        return MD4Hash(block_hashes).digest()

    def hexdigest(self, /) -> str:
        return self.digest().hex()


def ed2k_hash(file: Buffer | SupportsRead[bytes]) -> tuple[int, str]:
    if hasattr(file, "getbuffer"):
        file = file.getbuffer()
    if isinstance(file, Buffer):
        chunk_iter = bytes_to_chunk_iter(file, chunksize=MD4_BLOCK_SIZE)
    else:
        chunk_iter = bio_chunk_iter(file, chunksize=MD4_BLOCK_SIZE, can_buffer=True)
    block_hashes = bytearray()
    filesize = 0
    for chunk in chunk_iter:
        block_hashes += MD4Hash(chunk).digest()
        filesize += len(chunk)
    if not filesize % MD4_BLOCK_SIZE:
        block_hashes += MD4_EMPTY_HASH
    return filesize, MD4Hash(block_hashes).hexdigest()


async def ed2k_hash_async(file: Buffer | SupportsRead[bytes]) -> tuple[int, str]:
    if hasattr(file, "getbuffer"):
        file = file.getbuffer()
    if isinstance(file, Buffer):
        chunk_iter = bytes_to_chunk_async_iter(file, chunksize=MD4_BLOCK_SIZE)
    else:
        chunk_iter = bio_chunk_async_iter(file, chunksize=MD4_BLOCK_SIZE, can_buffer=True)
    block_hashes = bytearray()
    filesize = 0
    async for chunk in chunk_iter:
        block_hashes += MD4Hash(chunk).digest()
        filesize += len(chunk)
    if not filesize % MD4_BLOCK_SIZE:
        block_hashes += MD4_EMPTY_HASH
    return filesize, MD4Hash(block_hashes).hexdigest()




#!/usr/bin/env python3
# coding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__doc__ = "calculate file ed2k hash"

from argparse import ArgumentParser, RawTextHelpFormatter
from json import dumps
from os.path import basename, isfile

from iterdir import iterdir

parser = ArgumentParser(description=__doc__, formatter_class=RawTextHelpFormatter)
parser.add_argument("paths", metavar="path", nargs="*", help="file path(s) to be downloaded, if omitted, read from stdin (one path per line)")
parser.add_argument("-v", "--version", action="store_true", help="print the current version")

NAME_TRANSTAB = dict(zip(b"/|", ("%2F", "%7C")))


def parse_args(argv=None):
    args = parser.parse_args(argv)
    if args.version:
        from ed2k import __version__
        print(".".join(map(str, __version__)))
        raise SystemExit(0)
    return args

 
def main(argv=None):
    from ed2k import ed2k_hash 

    args = parse_args(argv)

    if args.paths:
        paths = args.paths
    else:
        from sys import stdin
        paths = (l.removesuffix("\n") for l in stdin)

    try:
        for path in paths:
            if not path:
                continue
            for file in iterdir(path, min_depth=0, max_depth=-1, predicate=lambda p: isfile(p)):
                size, ed2k = ed2k_hash(open(file, "rb"))
                print(dumps({
                    "path": file, 
                    "size": size, 
                    "ed2k": ed2k, 
                    "url": f"ed2k://|file|{basename(file).translate(NAME_TRANSTAB)}|{size}|{ed2k}|/", 
                }, ensure_ascii=False))
    except (EOFError, KeyboardInterrupt):
        pass
    except BrokenPipeError:
        from sys import stderr
        stderr.close()


if __name__ == "__main__":
    from pathlib import Path
    from sys import path

    path[0] = str(Path(__file__).parents[1])
    main()

